

/**
 ** sparkEnvironmentCodels.cc
 **
 ** Codels called by execution task sparkMainTask
 **
 ** Author: Emrah Akin Sisbot
 ** Date: Aug 2010
 **
 **/

#include <portLib.h>

#include "server/sparkHeader.h"
#include "sparkm3d.h"

#if defined(JIDO) || defined(HRP2)
#include "libeuler/pomEuler.h"
#include "sparkMocap.h"
#endif

#if defined(JIDO)
#include <niut/niutStruct.h>
#endif

// #ifdef ORO
// #include "sparkOro.h"
// #endif

/* ------------- Posters -------------- */

static POSTER_ID gVimanPosterID=NULL;  // Object position (Viman) poster
static POSTER_ID gSparkObjectPosterID=NULL; // Object position poster with spark_object_position type
static POSTER_ID gArmConfPosterID=NULL; // Arm or full robot configuration poster (depends on the robot)
static POSTER_ID gPosPosterID=NULL;     // Position (POM) poster
static POSTER_ID gPTPosterID=NULL;      // Pan tilt unit poster 
static POSTER_ID gRobotHandPosterID=NULL;    // Robot Hand configuration poster
static POSTER_ID gMocapPosterID=NULL;   // Motion capture poster 
static POSTER_ID gGestPosterID=NULL;    // Gest poster 
static POSTER_ID gNiutPosterID=NULL;    // Niut poster
static POSTER_ID gKinectPosterID=NULL;     // Kinect pos (POM)
static POSTER_ID gHumanConfPosterID=NULL;   // Morse Simulator poster 
static POSTER_ID gBodyPartsPosterID=NULL; // Spark yarp poster
static POSTER_ID gRobotConfPosterID=NULL;   // Configuration poster 

/* ----- Module internal variables with mostly Move3d types ----- */

static configPt gRRRobotQ = NULL;
static configPt gRRPrevRobotQ = NULL;

static configPt gRHHumanQ = NULL;

static SPARK_OBJECT_MOTION_HISTORY_ARRAY gObjHisArray;
static SPARK_AGENT_HISTORY *gHumanPrevPositions;
static int gHumanHeadJumpCounter[SPARK_MAX_AGENT_NB];
static int gHumanRHandJumpCounter[SPARK_MAX_AGENT_NB];
static int gHumanLHandJumpCounter[SPARK_MAX_AGENT_NB];
static int gHumanRHandReachFailCounter[SPARK_MAX_AGENT_NB];
static int gHumanLHandReachFailCounter[SPARK_MAX_AGENT_NB];

bool isKinectOnTable=false;
GEN_STRING64 PomKinectPoster = { "pomKinectFramePos" };

int place_robot_conf_to_environment_poster(p3d_rob* rob, configPt q)
{
  int i;
  int robIdx;
  
  if( (gEnv == NULL) || (rob == NULL) || (rob->num > SPARK_MAX_THINGS_NB-1) || (rob->nb_dof > M3D_MAX_DOF) )
    return FALSE;
  
  robIdx = rob->num;
  
  if(rob->nb_dof > M3D_MAX_DOF) {
    printf("%s has too many degrees of freedom (%d). Cannot write it into a poster\n",
	   rob->name, rob->nb_dof);
    return FALSE;
  }
  
  for(i=0; i<rob->nb_dof; i++) {
    SDI_F->envCur.robot[robIdx].q[i] = rob->ROBOT_POS[i] = q[i];
  }
  
  return TRUE;
}

/** 
    Added by Matthieu Warnier to update environment poster with p3d_rob of object. It will used when object is considered as being in hand and move with it. 
*/

int place_robot_p3d_rob_to_environment_poster(p3d_rob* rob)
{
  int i;
  int robIdx;
  
  if( (gEnv == NULL) || (rob == NULL) || (rob->num > SPARK_MAX_THINGS_NB-1) || (rob->nb_dof > M3D_MAX_DOF) )
    return FALSE;
  
  robIdx = rob->num;
  
  if(rob->nb_dof > M3D_MAX_DOF) {
    printf("%s has too many degrees of freedom (%d). Cannot write it into a poster\n",
	   rob->name, rob->nb_dof);
    return FALSE;
  }
  p3d_get_robot_config_into(rob,&rob->ROBOT_POS);
  for(i=0; i<rob->nb_dof; i++) {
    SDI_F->envCur.robot[robIdx].q[i] = rob->ROBOT_POS[i];
  }
  
  return TRUE;
}


void publish_environment_poster( void )
{
  SDI_F->envCur.time = time(0);
  
  sparkEnvironmentenvCurPosterWrite(SPARK_ENVIRONMENT_POSTER_ID, &SDI_F->envCur);
}

/* ------------------------------------------------------------- */
/* ----------------------- READ OBJECTS ------------------------ */
/* ------------------------------------------------------------- */

// static int update_object_motion(int sparkIdx, unsigned long updateTime, int motionState)
// {
//   int i,a_i;
//   HRI_KNOWLEDGE_ON_ENTITY * kn_on_ent;
//   HRI_AGENT * agent;

//   if(sparkIdx > SDI_F->objStates.length) {
//     printf("Given poster index is higher than the length.\n");
//     return FALSE;
//   }
  
//   if(updateTime != gObjHisArray.sensorTime[sparkIdx][0]) {
//     for(i=4; i>0; i--) {
//       gObjHisArray.motion[sparkIdx][i] = gObjHisArray.motion[sparkIdx][i-1];
//       gObjHisArray.sensorTime[sparkIdx][i] = gObjHisArray.sensorTime[sparkIdx][i-1];
//     }

//     // CURRENT READING
//     gObjHisArray.motion[sparkIdx][0] = (SPARK_OBJECT_MOTION)motionState;
//     gObjHisArray.sensorTime[sparkIdx][0] = updateTime;
//     SDI_F->objStates.object[sparkIdx].lastTimeSeen = updateTime;

//     if ( SDI_F->objStates.object[sparkIdx].detected == FALSE ) {
//       SDI_F->objStates.object[sparkIdx].detected = TRUE;
//       SDI_F->objStates.object[sparkIdx].exists = TRUE;
//       SDI_F->updateObjectsStates = TRUE;
//       printf("%s APPEARED\n", SDI_F->objStates.object[sparkIdx].name.name);
//       return TRUE;
//     }

//     // NEW STRUCTURE
//     if(gEntities->entities[sparkIdx]->disappeared) {
//       gEntities->entities[sparkIdx]->disappeared = FALSE;
//       // update disappeared_isexported for all agents to have isAt unknownLocation fact deleted in the ontology for all of them
//       for(a_i=0; a_i<gAgents->all_agents_no; a_i++){
// 	agent = gAgents->all_agents[a_i]; 
// 	kn_on_ent = &agent->knowledge->entities[sparkIdx]; 
// 	kn_on_ent->disappeared_isexported = FALSE;
//       }
//       SDI_F->updateObjectsStates = TRUE;
//       printf("%s APPEARED\n", SDI_F->objStates.object[sparkIdx].name.name);
//       return TRUE;
//     }
//     // END NEW STRUCTURE
    
    
//     // EVALUATE OBJECT MOTION
//     for(i=0; i<3; i++) {
//       if(gObjHisArray.motion[sparkIdx][i] == OBJ_MOVING)
// 	break;
//     }
    
//     if(i == 3) {
//       // OBJECT IS STATIC
//       if(SDI_F->objStates.object[sparkIdx].motion == OBJ_MOVING) {
//       	SDI_F->updateObjectsStates = TRUE;
// 	printf("%s HAS STOPPED\n", SDI_F->objStates.object[sparkIdx].name.name);
//       }
//       SDI_F->objStates.object[sparkIdx].lastTimeStatic = updateTime;
//       SDI_F->objStates.object[sparkIdx].motion = OBJ_STATIC;
//       // NEW STRUCTURE
//       if ( gRobot->knowledge->entities[sparkIdx].motion != HRI_STATIC) {
// 	gRobot->knowledge->entities[sparkIdx].motion_ischanged = TRUE;
// 	gRobot->knowledge->entities[sparkIdx].motion_isexported = FALSE;
// 	gRobot->knowledge->entities[sparkIdx].motion = HRI_STATIC;
//       }
//       // END NEW STRUCTURE
//     }
//     else {
//       if(SDI_F->objStates.object[sparkIdx].motion == OBJ_STATIC) {
// 	printf("%s BEGINS TO MOVE\n", SDI_F->objStates.object[sparkIdx].name.name);
// 	SDI_F->objStates.object[sparkIdx].motion = OBJ_MOVING;
//       	SDI_F->updateObjectsStates = TRUE;
//       }
//       printf("%s is MOVING\n", SDI_F->objStates.object[sparkIdx].name.name);
//       // NEW STRUCTURE
//       if ( gRobot->knowledge->entities[sparkIdx].motion != HRI_MOVING){
// 	gRobot->knowledge->entities[sparkIdx].motion = HRI_MOVING;
// 	gRobot->knowledge->entities[sparkIdx].motion_ischanged = TRUE;
// 	gRobot->knowledge->entities[sparkIdx].motion_isexported = FALSE;
//       }
//       // END NEW STRUCTURE      

//     }
//   } 
  
//   return TRUE;
// }


/* sparkReadObjectsStart  -  codel START of ReadObjects
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkReadObjectsStart(SPARK_INPUT_POSTER *objectPoster, int *report)
{
  if(gEnv == NULL) {
    printf("SPARK is not properly initialized.\n");
    *report = S_spark_NOT_INITIALIZED;
    return ETHER;
  }
  
  if(objectPoster->posterType == USE_VIMAN) {
    if(posterFind(objectPoster->name.name, &gVimanPosterID)==ERROR) {
      *report = S_spark_POSTER_NOT_FOUND;
	  printf("VIMAN poster not found!");
      SDI_F->state.vimanPosterAv = FALSE;
      return ETHER;
    }
    else {
      SDI_F->state.vimanPosterAv = TRUE;
      return EXEC;
    }
  }
  if(objectPoster->posterType == USE_OBJ_POS) {
    if(posterFind(objectPoster->name.name, &gSparkObjectPosterID)==ERROR) {
      *report = S_spark_POSTER_NOT_FOUND;
      SDI_F->state.sparkObjPosPosterAv = FALSE;
      return ETHER;
    }
    else {
      SDI_F->state.sparkObjPosPosterAv = TRUE;
      return EXEC;
    }
  }  
  if(objectPoster->posterType == USE_OBJ_INSIDE_SPARK) {
    SDI_F->state.useObjInsideSparkEnabled = TRUE;
    return EXEC;
  }

  SDI_F->state.vimanPosterAv = FALSE;
  SDI_F->state.sparkObjPosPosterAv = FALSE;
  SDI_F->state.useObjInsideSparkEnabled = FALSE;
  return ETHER;
}


/* sparkReadObjectsMain  -  codel EXEC of ReadObjects
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkReadObjectsMain(SPARK_INPUT_POSTER *objectPoster, int *report)
{
  int i, j, e_j,a_i;
  p3d_matrix4 posMat;
  double x, y, z, prev_x, prev_y, prev_z;
  configPt objectQ,robotQ;
  p3d_rob * rob = NULL;

  /* Spark object data structures */
  SPARK_OBJECT_POSITION_ARRAY objectPosPoster;

  HRI_KNOWLEDGE_ON_ENTITY * kn_on_ent;
  HRI_AGENT * agent;

#if defined(JIDO) || defined(HRP2)
  /* Viman data structures */
  VimanObjectPublicArray vimanPoster;
#endif

#if defined(JIDO) || defined(HRP2)

  if(SDI_F->state.vimanPosterAv) {
    if(posterRead(gVimanPosterID,0, &vimanPoster, sizeof(vimanPoster)) != sizeof(vimanPoster)) {
      *report = S_spark_POSTER_NOT_READ;
      return ETHER;
    }
   
    for(i=0; i<vimanPoster.nbObjects; i++) {
      for(e_j=0; e_j<gEntities->entities_nb; e_j++) {
        if (!strcmp(vimanPoster.objects[i].name, gEntities->entities[e_j]->robotPt->name)) {
	    break;
	}
      }
      
      // e_j is the index in entity list
      if (e_j == gEntities->entities_nb) {
	continue; // Viman Object is not found in SPARK
      }

      if (vimanPoster.objects[i].found_Stereo || vimanPoster.objects[i].found_Left || vimanPoster.objects[i].found_Right) {


	// Update Object Position - Only if found stereo to avoid localization problems.
	if (vimanPoster.objects[i].found_Stereo) {
	  prev_x = gEntities->entities[e_j]->robotPt->joints[1]->abs_pos[0][3];
	  prev_y = gEntities->entities[e_j]->robotPt->joints[1]->abs_pos[1][3];
	  prev_z = gEntities->entities[e_j]->robotPt->joints[1]->abs_pos[2][3];
	
	  x = vimanPoster.objects[i].thetaMatOrigin.px; // X
	  y = vimanPoster.objects[i].thetaMatOrigin.py; // Y
	  z = vimanPoster.objects[i].thetaMatOrigin.pz; // Z
	
	  if(D3D(x,y,z,prev_x,prev_y,prev_z) > SPARK_OBJECT_DEFAULT_UPDATE_DIST) {	
	  
	    objectQ =  MY_ALLOC(double, gEntities->entities[e_j]->robotPt->nb_dof); 
  
	    p3d_mat4Copy(p3d_mat4IDENTITY, posMat);
	    posMat[0][0] =  vimanPoster.objects[i].thetaMatOrigin.nx;
	    posMat[1][0] =  vimanPoster.objects[i].thetaMatOrigin.ny;
	    posMat[2][0] =  vimanPoster.objects[i].thetaMatOrigin.nz;
	  
	    posMat[0][1] =  vimanPoster.objects[i].thetaMatOrigin.ox;
	    posMat[1][1] =  vimanPoster.objects[i].thetaMatOrigin.oy;
	    posMat[2][1] =  vimanPoster.objects[i].thetaMatOrigin.oz;
	  
	    posMat[0][2] =  vimanPoster.objects[i].thetaMatOrigin.ax;
	    posMat[1][2] =  vimanPoster.objects[i].thetaMatOrigin.ay;
	    posMat[2][2] =  vimanPoster.objects[i].thetaMatOrigin.az;
	  
	    posMat[0][3] =  vimanPoster.objects[i].thetaMatOrigin.px; // X
	    posMat[1][3] =  vimanPoster.objects[i].thetaMatOrigin.py; // Y
	    posMat[2][3] =  vimanPoster.objects[i].thetaMatOrigin.pz; // Z
	  
	    p3d_set_freeflyer_pose_into(gEntities->entities[e_j]->robotPt, posMat, objectQ);
	  
	    p3d_set_and_update_this_robot_conf(gEntities->entities[e_j]->robotPt, objectQ);
	  
	    if(place_robot_conf_to_environment_poster(gEntities->entities[e_j]->robotPt, objectQ))
	      publish_environment_poster();
	  
	    //TODO: Check if mightabilities work with spark
	    //if(SDI_F->state.mightabilitiesEnabled)
	    //  execute_Mightability_Map_functions();
	    //Object position is updated in Move3D
	  }      
	}

	if(gEntities->entities[e_j]->detection_time != vimanPoster.objects[i].tacq_usec){
	  gEntities->entities[e_j]->is_present = TRUE;
	  gEntities->entities[e_j]->is_detected = TRUE;	

	  if(((HRI_MOTION) vimanPoster.objects[i].motion) == HRI_MOVING)
	    gEntities->entities[e_j]->last_ismoving_iter = SPARK_OBJECT_HISTORY_LENGTH; 
	}
	gEntities->entities[e_j]->last_detection_time = gEntities->entities[e_j]->detection_time; 
	gEntities->entities[e_j]->detection_time = vimanPoster.objects[i].tacq_usec;

	  //update_object_motion(e_j, vimanPoster.objects[i].tacq_usec, vimanPoster.objects[i].motion);
      } // endif viman detects stereo
      else {
	if(gEntities->entities[e_j]->undetection_status !=  HRI_NEVER_DETECTED)
	  gEntities->entities[e_j]->is_detected = FALSE;
	// if ( SDI_F->objStates.object[e_j].detected == TRUE ) {
	//   SDI_F->updateObjectsStates = TRUE; 
	//   printf("%s DISAPPEARED\n", SDI_F->objStates.object[e_j].name.name);
	// }
	// SDI_F->objStates.object[e_j].detected = FALSE;
	
	// New Structures
	// if( gEntities->entities[e_j]->is_present ) {
        //   if(!gEntities->entities[e_j]->disappeared){
	//     printf("%s DISAPPEARED\n", SDI_F->objStates.object[e_j].name.name);
	//     gEntities->entities[e_j]->disappeared = TRUE;
	//     // update disappeared_isexported for all agents to have facts deleted in the ontology for all of them
	//     for(a_i=0; a_i<gAgents->all_agents_no; a_i++){
	//       agent = gAgents->all_agents[a_i]; 
	//       kn_on_ent = &agent->knowledge->entities[e_j]; 
	//       kn_on_ent->disappeared_isexported = FALSE;
	//     }

	//     // put object in 0,0,0 if disappear. First draft need to be improved depending on viisibility placement and isvisible for robot.	
	   
	//     //objectQ = MY_ALLOC(double, gEntities->entities[e_j]->robotPt->nb_dof); /* ALLOC */
	//     //p3d_get_robot_config_into(gEntities->entities[e_j]->robotPt, &objectQ);
	//     //objectQ[6] = objectQ[7] = objectQ[8] = 0;      
	//     //p3d_set_and_update_this_robot_conf(gEntities->entities[e_j]->robotPt, objectQ);
	//     //MY_FREE(objectQ, double, gEntities->entities[e_j]->robotPt->nb_dof); /* FREE */
	//   }
	  
	//   SDI_F->updateObjectsStates = TRUE;
	// }
	// End New Structures
	
      } 
    } // endfor vimanposter objects
  }

  // simulation of obj import using obj position inside spark that can be moved by ( PlaceObject )
  if(SDI_F->state.useObjInsideSparkEnabled) {
    
    // //for(i=0; i<vimanPoster.nbObjects; i++) {
    // if (SDI_F->addedObjectSimu.isnew) {
    //   //we need to manage several times the object to manage is moving properly
    //   if ( SDI_F->addedObjectSimu.isnew < 7 )
    // 	SDI_F->addedObjectSimu.isnew++;
    //   else
    // 	SDI_F->addedObjectSimu.isnew = 0;
    for(i=0; i<gEnv->nr; i++) {
      if( strcasestr(gEnv->robot[i]->name, SDI_F->addedObjectSimu.name.name)) {
	rob = gEnv->robot[i];
	break;
      }
    }
      
    for(e_j=0; e_j<gEntities->entities_nb; e_j++) {
      if (!strcmp(gEnv->robot[i]->name, gEntities->entities[e_j]->robotPt->name)){
	// printf("Env Entity %s %s\n", gEnv->robot[i]->name, gEntities->entities[e_j]->robotPt->name);
	break;
      }
    }
      
    // e_j is the index in entity list
    if (e_j < gEntities->entities_nb) {
      // Something should be done to mnanage disappear for simulation.
      if (!SDI_F->addedObjectSimu.hasDisappeared) {

	gEntities->entities[e_j]->is_present = TRUE;

	// Update Object Position
	prev_x = gEntities->entities[e_j]->robotPt->joints[1]->abs_pos[0][3];
	prev_y = gEntities->entities[e_j]->robotPt->joints[1]->abs_pos[1][3];
	prev_z = gEntities->entities[e_j]->robotPt->joints[1]->abs_pos[2][3];
	
	x = SDI_F->addedObjectSimu.coords.x; // X
	y = SDI_F->addedObjectSimu.coords.y; // Y
	z = SDI_F->addedObjectSimu.coords.z;  // Z
	//printf("%f %f %f %f %f %f\n",prev_x,prev_y,prev_z,x,y,z);
	if(D3D(x,y,z,prev_x,prev_y,prev_z) > SPARK_OBJECT_DEFAULT_UPDATE_DIST) {	
	  rob = gEnv->robot[i];
	  robotQ = MY_ALLOC(double, rob->nb_dof); /* ALLOC */
	  p3d_get_robot_config_into(rob, &robotQ);
	  
	  robotQ[6]  = SDI_F->addedObjectSimu.coords.x;
	  robotQ[7]  = SDI_F->addedObjectSimu.coords.y;
	  robotQ[8]  = SDI_F->addedObjectSimu.coords.z;
	  robotQ[9]  = DTOR(SDI_F->addedObjectSimu.coords.yaw);
	  robotQ[10] = DTOR(SDI_F->addedObjectSimu.coords.pitch);
	  robotQ[11] = DTOR(SDI_F->addedObjectSimu.coords.roll);
  
	  p3d_set_and_update_this_robot_conf(rob, robotQ);
  
	  if(place_robot_conf_to_environment_poster(rob, robotQ))
	    publish_environment_poster();

	  MY_FREE(robotQ, double, rob->nb_dof); /* FREE */
	  //update_object_motion(e_j,clock(), OBJ_MOVING);
	  gEntities->entities[e_j]->last_ismoving_iter = SPARK_OBJECT_HISTORY_LENGTH;
	}

	gEntities->entities[e_j]->is_present = TRUE;
	gEntities->entities[e_j]->is_detected = TRUE;			 
	  
	gEntities->entities[e_j]->last_detection_time = gEntities->entities[e_j]->detection_time; 
	gEntities->entities[e_j]->detection_time = clock();


      } // endif hasDisappeared false.
      else {
	if(gEntities->entities[e_j]->undetection_status !=  HRI_NEVER_DETECTED)
	  gEntities->entities[e_j]->is_detected = FALSE;
	// if ( SDI_F->objStates.object[e_j].detected == TRUE ) {
	//   SDI_F->updateObjectsStates = TRUE; 
	//   printf("%s DISAPPEARED\n", SDI_F->objStates.object[e_j].name.name);
	// }
	// SDI_F->objStates.object[e_j].detected = FALSE;
	
	// // New Structures
	// if( gEntities->entities[e_j]->is_present ) {
	//   if(!gEntities->entities[e_j]->disappeared){
	//     printf("%s DISAPPEARED\n", SDI_F->objStates.object[e_j].name.name);
	    
	//     //update and export disappear only once
	//     if ( SDI_F->addedObjectSimu.isnew == 2 ) {
	// 	gEntities->entities[e_j]->disappeared = TRUE;
	// 	// update disappeared_isexported for all agents to have facts deleted in the ontology for all of them
	// 	for(a_i=0; a_i<gAgents->all_agents_no; a_i++) { 
	// 	  agent = gAgents->all_agents[a_i]; 
	// 	  kn_on_ent = &agent->knowledge->entities[e_j]; 
	// 	  kn_on_ent->disappeared_isexported = FALSE;
	// 	}

	// 	// put object in 0,0,0 if disappear. First draft need to be improved depending on viisibility placement and isvisible for robot.	
	// 	objectQ = MY_ALLOC(double, gEnv->robot[i]->nb_dof); /* ALLOC */
	// 	p3d_get_robot_config_into(gEnv->robot[i], &objectQ);
	// 	objectQ[6] = objectQ[7] = objectQ[8] = 0;      
	// 	p3d_set_and_update_this_robot_conf(gEnv->robot[i], objectQ);
	// 	MY_FREE(objectQ, double, gEnv->robot[i]->nb_dof); /* FREE */
	//     }
	//   }
	//   SDI_F->updateObjectsStates = TRUE;
	// }
	// End New Structures	
      }
    }
  } // if new sim

  

#endif
  if(SDI_F->state.sparkObjPosPosterAv) {
    if(posterRead(gSparkObjectPosterID, 0, &objectPosPoster, sizeof(objectPosPoster)) != sizeof(objectPosPoster)) {
      *report = S_spark_POSTER_NOT_READ;
      return ETHER;
    }
    
    for(i=0; i<objectPosPoster.nbObjects; i++) {
      for(j=0; j<gEnv->nr; j++) {
	if (!strcmp(objectPosPoster.objects[i].name.name, gEnv->robot[j]->name)) {
	  break;
	}
      }
      
      // j is the index in Move3D
      if (j == gEnv->nr) {
	continue; // Object is not found in SPARK
      }
      
      if (objectPosPoster.objects[i].visible) {
	// Update Object Position
	prev_x = gEnv->robot[j]->joints[1]->abs_pos[0][3];
 	prev_y = gEnv->robot[j]->joints[1]->abs_pos[1][3];
	prev_z = gEnv->robot[j]->joints[1]->abs_pos[2][3];
	
	x = objectPosPoster.objects[i].absPosMatrix.px; // X
	y = objectPosPoster.objects[i].absPosMatrix.py; // Y
	z = objectPosPoster.objects[i].absPosMatrix.pz; // Z
       
	//printf("%s position %f %f %f\n", gEnv->robot[j]->name, x, y, z);

	if(D3D(x,y,z,prev_x,prev_y,prev_z) > SPARK_OBJECT_DEFAULT_UPDATE_DIST) {	
	  
	  objectQ =  MY_ALLOC(double, gEnv->robot[j]->nb_dof); 
  
	  p3d_mat4Copy(p3d_mat4IDENTITY, posMat);
	  posMat[0][0] =  objectPosPoster.objects[i].absPosMatrix.nx;
	  posMat[1][0] =  objectPosPoster.objects[i].absPosMatrix.ny;
	  posMat[2][0] =  objectPosPoster.objects[i].absPosMatrix.nz;
	  
	  posMat[0][1] =  objectPosPoster.objects[i].absPosMatrix.ox;
	  posMat[1][1] =  objectPosPoster.objects[i].absPosMatrix.oy;
	  posMat[2][1] =  objectPosPoster.objects[i].absPosMatrix.oz;
	  
	  posMat[0][2] =  objectPosPoster.objects[i].absPosMatrix.ax;
	  posMat[1][2] =  objectPosPoster.objects[i].absPosMatrix.ay;
	  posMat[2][2] =  objectPosPoster.objects[i].absPosMatrix.az;
	  
	  posMat[0][3] =  objectPosPoster.objects[i].absPosMatrix.px; // X
	  posMat[1][3] =  objectPosPoster.objects[i].absPosMatrix.py; // Y
	  posMat[2][3] =  objectPosPoster.objects[i].absPosMatrix.pz; // Z
	  
	  p3d_set_freeflyer_pose_into(gEnv->robot[j], posMat, objectQ);
	  
	  p3d_set_and_update_this_robot_conf(gEnv->robot[j], objectQ);
	  
	  if(place_robot_conf_to_environment_poster(gEnv->robot[j], objectQ))
	    publish_environment_poster();
	  
	  //TODO: Check if mightabilities work with spark
	  //if(SDI_F->state.mightabilitiesEnabled)
	  //  execute_Mightability_Map_functions();
	  //Object position is updated in Move3D
	}

	if(gEntities->entities[e_j]->detection_time != objectPosPoster.objects[i].tacq_usec){
	  gEntities->entities[e_j]->is_present = TRUE;
	  gEntities->entities[e_j]->is_detected = TRUE;	

	  if(((HRI_MOTION) objectPosPoster.objects[i].motion) == HRI_MOVING)
	    gEntities->entities[e_j]->last_ismoving_iter = SPARK_OBJECT_HISTORY_LENGTH; 
	}
	gEntities->entities[e_j]->last_detection_time = gEntities->entities[e_j]->detection_time; 
	gEntities->entities[e_j]->detection_time = objectPosPoster.objects[i].tacq_usec;
	
	//update_object_motion(j, objectPosPoster.objects[i].tacq_usec, objectPosPoster.objects[i].motion);


      } // endif invisible
      else {
	if(gEntities->entities[e_j]->undetection_status !=  HRI_NEVER_DETECTED)
	  gEntities->entities[e_j]->is_detected = FALSE;

	// if ( SDI_F->objStates.object[j].detected == TRUE ) {
	//   SDI_F->updateObjectsStates = TRUE; 
	//   printf("OBJECT DISAPPEARED\n");
	// }
	// SDI_F->objStates.object[j].detected = FALSE;

      }
    } // endfor objectPosPoster objects
  }
  
  // update also oro here to put ismoving facts in oro not only in compute facts
// #ifdef ORO
//   prepareFactsForORO(gAgents, gEntities);
//   sendToORO(SDI_F->oroServerAddress.address, SDI_F->oroServerAddress.port);
// #endif

  return EXEC;
}


/* ------------------------------------------------------------- */
/* ----------------------- READ ROBOT -------------------------- */
/* ------------------------------------------------------------- */

#ifdef HRP2
static void Arm2ConfigPt(ghrp2_config_t *cfg, configPt m3dconfig)
{
   int i;
   
  for(i=0; i<16; i++)
    m3dconfig[12+i] = cfg->angles[i];
  
  for(i=0; i<7; i++)
    m3dconfig[29+i] = cfg->angles[GHRP2_RARM_JOINT0+i];
  
  for(i=0; i<7; i++)
    m3dconfig[42+i] = cfg->angles[GHRP2_LARM_JOINT0+i];
}
#elif defined(JIDO)
static void Arm2ConfigPt(Gb_q7 *gbconfig, configPt m3dconfig)
{
  m3dconfig[14] = gbconfig->q1;
  m3dconfig[15] = gbconfig->q2; 
  m3dconfig[16] = gbconfig->q3; 
  m3dconfig[17] = gbconfig->q4;
  m3dconfig[18] = gbconfig->q5; 
  m3dconfig[19] = gbconfig->q6;
  m3dconfig[20] = gbconfig->q7;
}

static void SAHand2ConfigPt(GENMANIP_SAHAND_STR *handCfg, configPt m3dconfig)
{
  if(handCfg!=NULL) {
    m3dconfig[21] = 1.57;//handCfg->handRight.thumbRotation;
    m3dconfig[22] = handCfg->handRight.thumb.base;
    m3dconfig[23] = handCfg->handRight.thumb.phaProx;
    m3dconfig[24] = handCfg->handRight.thumb.phaMid;
    m3dconfig[25] = handCfg->handRight.thumb.phaDist;
    
    m3dconfig[26] = handCfg->handRight.fore.base;
    m3dconfig[27] = handCfg->handRight.fore.phaProx;
    m3dconfig[28] = handCfg->handRight.fore.phaMid;
    m3dconfig[29] = handCfg->handRight.fore.phaDist;
    
    m3dconfig[30] = handCfg->handRight.middle.base;
    m3dconfig[31] = handCfg->handRight.middle.phaProx;
    m3dconfig[32] = handCfg->handRight.middle.phaMid;
    m3dconfig[33] = handCfg->handRight.middle.phaDist;
    
    m3dconfig[34] = handCfg->handRight.ring.base;
    m3dconfig[35] = handCfg->handRight.ring.phaProx;
    m3dconfig[36] = handCfg->handRight.ring.phaMid;
    m3dconfig[37] = handCfg->handRight.ring.phaDist;
  }
}
#endif

static void Pt2ConfigPt(double pan, double tilt,  configPt m3dconfig)
{
  m3dconfig[12]  = pan; 
  m3dconfig[13]  = -tilt;
}


/* sparkReadRobotStart  -  codel START of ReadRobot
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkReadRobotStart(SPARK_ROBOT_POSTERS *robotPosters, int *report)
{
  if(gEnv == NULL) {
    printf("SPARK is not properly initialized.\n");
    *report = S_spark_NOT_INITIALIZED;
    return ETHER;
  }

  if(posterFind(robotPosters->armConfPostername.name, &gArmConfPosterID) == ERROR) {
    printf(" Can't find %s Poster\n",robotPosters->armConfPostername.name);
    SDI_F->state.armConfPosterAv = FALSE;
  }
  else {
    SDI_F->state.armConfPosterAv = TRUE;
  }
  if(posterFind(robotPosters->handConfPostername.name, &gRobotHandPosterID) == ERROR) {
    printf(" Can't find %s Poster\n",robotPosters->handConfPostername.name);
    SDI_F->state.robotHandPosterAv = FALSE;
  }
  else {
    SDI_F->state.robotHandPosterAv = TRUE;
  }
  if(posterFind(robotPosters->posPostername.name, &gPosPosterID) == ERROR) {
    printf(" Can't find %s Poster\n",robotPosters->posPostername.name);
    SDI_F->state.posPosterAv = FALSE;
  }
  else {
    SDI_F->state.posPosterAv = TRUE;
  }
  
  if(posterFind(robotPosters->ptPostername.name, &gPTPosterID) == ERROR) {
    printf(" Can't find %s Poster\n",robotPosters->ptPostername.name);
    SDI_F->state.ptPosterAv = FALSE;
  }
  else {
    SDI_F->state.ptPosterAv = TRUE;
  }
  
#if !defined(JIDO)   // TODO: There is bug somewhere which bloks in posterfind when a poster doesn't exist.
  if(posterFind(robotPosters->confPosterName.name, &gRobotConfPosterID) == ERROR) {
    printf(" Can't find %s Poster\n", robotPosters->confPosterName.name);
    SDI_F->state.robotConfPosterAv = FALSE;
  }
  else {
    SDI_F->state.robotConfPosterAv = TRUE;
  }
#else
  SDI_F->state.robotConfPosterAv = FALSE;
#endif

  gRRRobotQ = MY_ALLOC(double, gRobot->robotPt->nb_dof); /* Allocation of temporary robot configuration */
  gRRPrevRobotQ = MY_ALLOC(double, gRobot->robotPt->nb_dof); /* Allocation of temporary robot configuration */
  
  return EXEC;
}


/* sparkReadRobotMain  -  codel EXEC of ReadRobot
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkReadRobotMain(SPARK_ROBOT_POSTERS *robotPosters, int *report)
{
#if defined(HRP2)
  ghrp2_config_t cfg;
  POM_POS basePos;
#elif defined(JIDO)
  Gb_q7 cfg;
  POM_SENSOR_POS platineCfg;
  T3D platineFrame;
  POM_POS basePos;
  GENMANIP_SAHAND_STR handCfg;
#endif
  double eefPos[3];
  int panQIdx, tiltQIdx;
  SPARK_CONFIGURATION conf;
  int i;
  
  // copy current configuration
  p3d_get_robot_config_into(gRobot->robotPt, &gRRRobotQ);
  p3d_get_robot_config_into(gRobot->robotPt, &gRRPrevRobotQ);
  
  if(SDI_F->state.robotConfPosterAv) {
    if (posterRead(gRobotConfPosterID,0, &conf, sizeof(conf)) != sizeof(conf)) {
      *report = S_spark_POSTER_NOT_READ;
      return END;
    }
    // Here we consider that the configuration poster is well built
    // Meaning that the indexes and the order of joint angles are the same in the poster and in move3d
    // We apply the data directly without any test (we cannot test anyways)
    for(i=0; i<conf.dofNb; i++) {
      gRRRobotQ[i] = conf.dof[i];
    }
  }
#if defined(HRP2) || defined(JIDO) 
 
  if(SDI_F->state.armConfPosterAv) {
    if (posterRead(gArmConfPosterID,0, &cfg, sizeof(cfg)) != sizeof(cfg)) {
      *report = S_spark_POSTER_NOT_READ;
      return END;
    }
    // copy new arm position into configuration
    Arm2ConfigPt(&cfg, gRRRobotQ);
  }
  
  if(SDI_F->state.posPosterAv) {
    if (posterRead(gPosPosterID,0, &basePos, sizeof(basePos)) != sizeof(basePos)) {
      *report =  S_spark_POSTER_NOT_READ;
      return END;
    }
    gRRRobotQ[6]  = basePos.mainToOrigin.euler.x;
    gRRRobotQ[7]  = basePos.mainToOrigin.euler.y;
    gRRRobotQ[8]  = basePos.mainToOrigin.euler.z;
    gRRRobotQ[11] = basePos.mainToOrigin.euler.yaw;
  }
  
  if(SDI_F->state.ptPosterAv) {
    if (posterRead(gPTPosterID, 0, &platineCfg, sizeof(platineCfg)) != sizeof(platineCfg)) {
      *report =  S_spark_POSTER_NOT_READ;
      return END;
    }
    
    pomEulerVToT3D(&(platineCfg.sensorToMain), &platineFrame);
    platineFrame.euler.flags |= T3D_ALLOW_CONVERSION;
    t3dConvertTo(T3D_EULER, &platineFrame);
    // copy head platine position to config
    Pt2ConfigPt(platineFrame.euler.euler[0], platineFrame.euler.euler[1], gRRRobotQ);
  }
  
  if(SDI_F->state.robotHandPosterAv) {
    if (posterRead(gRobotHandPosterID, 0, &handCfg, sizeof(handCfg)) != sizeof(handCfg)) {
      *report =  S_spark_POSTER_NOT_READ;
      return END;
    }
    // copy hand configuration to config
    SAHand2ConfigPt(&handCfg, gRRRobotQ);
  }

#endif
  // Test if the object states need to be updated
  panQIdx  = gRobot->robotPt->joints[gRobot->perspective->pan_jnt_idx]->index_dof;
  tiltQIdx = gRobot->robotPt->joints[gRobot->perspective->tilt_jnt_idx]->index_dof;
  
  if( D3D(gRRRobotQ[6], gRRRobotQ[7], gRRRobotQ[8],
	  gRRPrevRobotQ[6], gRRPrevRobotQ[7], gRRPrevRobotQ[8]) > 0.1 ) { // if the displacement is greater than 10cm
    SDI_F->updateObjectsStates = TRUE;
    printf("ROBOT BASE MOVED\n");
  }
  //  else {
  //     if( D2D(gRRRobotQ[panQIdx], gRRRobotQ[tiltQIdx], 
  // 	    gRRPrevRobotQ[panQIdx], gRRPrevRobotQ[tiltQIdx]) > 0.1 ) { // if the platine changes is greater than 0.1rad
  //       SDI_F->updateObjectsStates = TRUE;
  //        printf("ROBOT platine MOVED\n");
  //     }
  else {
    eefPos[0] = gRobot->perspective->pointjoint->abs_pos[0][3];
    eefPos[1] = gRobot->perspective->pointjoint->abs_pos[1][3];
    eefPos[2] = gRobot->perspective->pointjoint->abs_pos[2][3];
  }
  //}

  // update robot pose in move3d
  p3d_set_and_update_this_robot_conf(gRobot->robotPt, gRRRobotQ);

  // Computes and exposes the minimal distance and cost
  // between the robot and all the humans in the scene
  SDI_F->distToHuman.costDistRobToHum = hri_distance_cost( gAgents, SDI_F->distToHuman.minDistRobToHum );
  sparkDistanceToHumandistToHumanPosterWrite(SPARK_DISTANCETOHUMAN_POSTER_ID, &SDI_F->distToHuman);
    
  //TODO: Check migthabilities-spark compatibility
  //if(SDI_F->state.mightabilitiesEnabled)
  //  execute_Mightability_Map_functions();

  if(!SDI_F->updateObjectsStates) {
    if( D3D(eefPos[0],eefPos[1],eefPos[2],
	    gRobot->perspective->pointjoint->abs_pos[0][3],
	    gRobot->perspective->pointjoint->abs_pos[1][3],
	    gRobot->perspective->pointjoint->abs_pos[2][3]) > 0.1 ) {
      SDI_F->updateObjectsStates = TRUE;
        printf("ROBOT eef MOVED\n");
      if(gRobot->perspective->enable_visible_objects_draw) {
	hri_compute_agent_sees(gRobot, 50, FALSE, FALSE);
      }
    }
  }
  
  if(place_robot_conf_to_environment_poster(gRobot->robotPt, gRRRobotQ))
    publish_environment_poster();

  return EXEC;
}

/* sparkReadRobotEnd  -  codel END of ReadRobot
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkReadRobotEnd(SPARK_ROBOT_POSTERS *robotPosters, int *report)
{
  MY_FREE(gRRRobotQ, double, gRobot->robotPt->nb_dof);  /* FREE */
  MY_FREE(gRRPrevRobotQ, double, gRobot->robotPt->nb_dof);  /* FREE */
  printf("Robot Synchonization Stopped.\n");
  *report = OK;
  return ETHER;
}


/* sparkReadRobotInter  -  codel INTER of ReadRobot
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkReadRobotInter(SPARK_ROBOT_POSTERS *robotPosters, int *report)
{
  MY_FREE(gRRRobotQ, double, gRobot->robotPt->nb_dof);  /* FREE */
  MY_FREE(gRRPrevRobotQ, double, gRobot->robotPt->nb_dof);  /* FREE */
  printf("Robot Synchonization Interrupted.\n");
  *report = OK;
  return ETHER;
}


/* ------------------------------------------------------------- */
/* ----------------------- READ HUMANS-------------------------- */
/* ------------------------------------------------------------- */

SPARK_AGENT_HISTORY * create_initialize_human_history(int humanNo)
{
  int i,j;
  SPARK_AGENT_HISTORY *historyForAll;
  
  historyForAll = MY_ALLOC(SPARK_AGENT_HISTORY, humanNo); 
  
  for(i=0; i<humanNo ; i++) { 
    for(j=0; j<SPARK_HUMAN_HISTORY_LENGTH; j++) {
      historyForAll[i].history[j].head.x = 0;
      historyForAll[i].history[j].head.y = 0;
      historyForAll[i].history[j].head.z = 0;
      historyForAll[i].history[j].head.roll = 0;
      historyForAll[i].history[j].head.pitch = 0;
      historyForAll[i].history[j].head.yaw = 0;
      
      historyForAll[i].history[j].lhand.x = 0;
      historyForAll[i].history[j].lhand.y = 0;
      historyForAll[i].history[j].lhand.z = 0;
      
      historyForAll[i].history[j].rhand.x = 0;
      historyForAll[i].history[j].rhand.y = 0;
      historyForAll[i].history[j].rhand.z = 0;
    }
    historyForAll[i].length = 0;
  }
  
  return historyForAll;
}

static int spark_push_position_to_history(SPARK_AGENT_HISTORY *history, int agentNo, SPARK_6D_COORD *head, SPARK_3D_COORD *lhand, SPARK_3D_COORD *rhand)
{
  int i;
  
  if(head==NULL && lhand==NULL && rhand==NULL) 
    return FALSE;
  
  if(head!=NULL) {
    //first shift array values to bottom
    for (i=SPARK_HUMAN_HISTORY_LENGTH-1; i>1; i--) {
      history[agentNo].history[i].head.x     = history[agentNo].history[i-1].head.x;
      history[agentNo].history[i].head.y     = history[agentNo].history[i-1].head.y;
      history[agentNo].history[i].head.z     = history[agentNo].history[i-1].head.z;
      history[agentNo].history[i].head.roll  = history[agentNo].history[i-1].head.roll;
      history[agentNo].history[i].head.pitch = history[agentNo].history[i-1].head.pitch;
      history[agentNo].history[i].head.yaw   = history[agentNo].history[i-1].head.yaw;
      history[agentNo].history[i].lhand.x    = history[agentNo].history[i-1].lhand.x;
      history[agentNo].history[i].lhand.y    = history[agentNo].history[i-1].lhand.y;
      history[agentNo].history[i].lhand.z    = history[agentNo].history[i-1].lhand.z;
      history[agentNo].history[i].rhand.x    = history[agentNo].history[i-1].rhand.x;
      history[agentNo].history[i].rhand.y    = history[agentNo].history[i-1].rhand.y;
      history[agentNo].history[i].rhand.z    = history[agentNo].history[i-1].rhand.z;
    }
    // put current values to the top of the pile
    history[agentNo].history[0].head.x     = head->x;
    history[agentNo].history[0].head.y     = head->y;
    history[agentNo].history[0].head.z     = head->z;
    history[agentNo].history[0].head.yaw   = head->yaw;
    history[agentNo].history[0].head.pitch = head->pitch;
    history[agentNo].history[0].lhand.x    = history[agentNo].history[1].lhand.x;
    history[agentNo].history[0].lhand.y    = history[agentNo].history[1].lhand.y;
    history[agentNo].history[0].lhand.z    = history[agentNo].history[1].lhand.z;
    history[agentNo].history[0].rhand.x    = history[agentNo].history[1].rhand.x;
    history[agentNo].history[0].rhand.y    = history[agentNo].history[1].rhand.y;
    history[agentNo].history[0].rhand.z    = history[agentNo].history[1].rhand.z;
    history[agentNo].length = MIN( history[agentNo].length+1, SPARK_HUMAN_HISTORY_LENGTH);
  }
  
  if(lhand!=NULL) {
    // put current values to the top of the pile
    history[agentNo].history[0].lhand.x = lhand->x;
    history[agentNo].history[0].lhand.y = lhand->y;
    history[agentNo].history[0].lhand.z = lhand->z;
  }
  
  if(rhand!=NULL) {
    // put current values to the top of the pile
    history[agentNo].history[0].rhand.x = rhand->x;
    history[agentNo].history[0].rhand.y = rhand->y;
    history[agentNo].history[0].rhand.z = rhand->z;
  }
  
  return TRUE;  
}

/* sparkReadHumansStart  -  codel START of ReadHumans
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkReadHumansStart(SPARK_INPUT_POSTER *humanPoster, int *report)
{
  int i;

  if(gAgents == NULL) {
    *report = S_spark_NOT_INITIALIZED;
    return ETHER;
  }
 
  gHumanPrevPositions = create_initialize_human_history(gAgents->humans_no);
  
  for(i=0; i<SPARK_MAX_AGENT_NB; i++) {
    gHumanHeadJumpCounter[i] = 0;
    gHumanRHandJumpCounter[i] = 0;
    gHumanLHandJumpCounter[i] = 0;
    gHumanRHandReachFailCounter[i] = 0;
    gHumanLHandReachFailCounter[i] = 0;
  }
  
  if(humanPoster->posterType == USE_MOCAP) {    
    if(posterFind( humanPoster->name.name, &gMocapPosterID) == ERROR) {
      *report = S_spark_POSTER_NOT_FOUND;
      SDI_F->state.mocapPosterAv = FALSE;
      return ETHER;
    }
    else {
      SDI_F->state.mocapPosterAv = TRUE;
      gRHHumanQ = MY_ALLOC(double, gAgents->humans[0]->robotPt->nb_dof); /* ALLOC */
      return EXEC;
    }
  }
  
#ifdef GEST_HUMAN_TRACK
  if(humanPoster->posterType == USE_GEST) {
    if(posterFind( humanPoster->name.name, &gGestPosterID) == ERROR) {
      *report = S_spark_POSTER_NOT_FOUND;
      SDI_F->state.handPosterAv = FALSE;
      return ETHER;
    }
    else {
      SDI_F->state.handPosterAv = TRUE;      
      gRHHumanQ = MY_ALLOC(double, gAgents->humans[0]->robotPt->nb_dof); /* ALLOC */
      return EXEC;
    }
  }
#endif

  //-----------------------------------------------------------------
  if(humanPoster->posterType == USE_NIUT) 
    {
      if(posterFind( humanPoster->name.name, &gNiutPosterID) == ERROR) {
	*report = S_spark_POSTER_NOT_FOUND;
	SDI_F->state.niutPosterAv = FALSE;
	return ETHER;
      }
      else {
	SDI_F->state.niutPosterAv = TRUE;      
	gRHHumanQ = MY_ALLOC(double, gAgents->humans[0]->robotPt->nb_dof); /* ALLOC */
      }
      // If kinect is on table we don't need
      // the POM poster
      if(isKinectOnTable) {
	return EXEC;
      }
      if(posterFind( PomKinectPoster.name , &gKinectPosterID) == ERROR) {
	*report = S_spark_POSTER_NOT_FOUND;
	SDI_F->state.kinectPosterAv = FALSE;
	return ETHER;
      }
      else {
	SDI_F->state.kinectPosterAv = TRUE;      
	*report = S_spark_POSTER_NOT_FOUND;
	return EXEC;
      }
    }
  //-----------------------------------------------------------------
  
  if(humanPoster->posterType == USE_CONF) {    
    if(posterFind( humanPoster->name.name, &gHumanConfPosterID) == ERROR) {
      *report = S_spark_POSTER_NOT_FOUND;
      SDI_F->state.humanConfPosterAv = FALSE;
      return ETHER;
    }
    else {
	  printf("Found a full human configuration poster. Using it.");
      SDI_F->state.humanConfPosterAv = TRUE;
      gRHHumanQ = MY_ALLOC(double, gAgents->humans[0]->robotPt->nb_dof); /* ALLOC */
      return EXEC;
    }
  }

  if(humanPoster->posterType == USE_BODY_PARTS) {    
    if(posterFind( humanPoster->name.name, &gBodyPartsPosterID) == ERROR) {
      *report = S_spark_POSTER_NOT_FOUND;
      SDI_F->state.bodyPartsPosterAv = FALSE;
      return ETHER;
    }
    else {
      SDI_F->state.bodyPartsPosterAv = TRUE;
      gRHHumanQ = MY_ALLOC(double, gAgents->humans[0]->robotPt->nb_dof); /* ALLOC */
      return EXEC;
    }
  } 
  
  SDI_F->state.mocapPosterAv = FALSE;
  SDI_F->state.handPosterAv = FALSE;
  SDI_F->state.humanConfPosterAv = FALSE;
  SDI_F->state.bodyPartsPosterAv = FALSE;
  
  *report = S_spark_POSTER_NOT_FOUND;
  return ETHER;
}


#if defined(JIDO)
void niutGenPosToVect(p3d_matrix4 mat, p3d_vector3 result, GEN_POINT_3D pos)
{
  p3d_vector3 vect;

  vect[0] = pos.x;
  vect[1] = pos.y;
  vect[2] = pos.z;

  p3d_xformPoint(mat,vect,result);
}
#endif

/* sparkReadHumansMain  -  codel EXEC of ReadHumans
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkReadHumansMain(SPARK_INPUT_POSTER *humanPoster, int *report)
{
  SPARK_AGENT_SENSOR_DATA humanRawPositions[gAgents->humans_no];
  int humanNb = 0;
  p3d_vector3 tCoord;
  int headDetected = FALSE;
  int i, j, hIdx;
  int lookingAtRobot = FALSE;
  int updateHumanPosture = FALSE;
  int headOrientationAvailable = FALSE;
  int headIsUpdated = FALSE;
  double updateDistDiff = 0;
  double updateAngleDiff = 0;
  int panDofIdx, tiltDofIdx;
  SPARK_HUMAN_STATE_ENUM state;
#if defined(JIDO) || defined(HRP2)
  T3D gestSensorMatrix;
  /* Mocap variables */
  MOCAP_OBJECTS mocapObjects;
  /* Gest variables */
#if defined(GEST_HUMAN_TRACK)
  GEST_POSITION_POSTER_STR gestPoster;
#endif
  POM_EULER_V cameraToOrigin;
#endif

#if defined(JIDO)
  /* Niut variables */ 
  NIUT_HUMAN_LIST niutPoster;
  POM_SENSOR_POS  kinectCfg;
  T3D kinectToRobot,robotToOrigin;
#endif

  /* Full configuration variables */
  SPARK_CONFIGURATION humanConf;
  /* Bodyparts variables */
  SPARK_HUMAN_BODY_PARTS_LIST humanBodyParts;
 
  for(i=0; i<gAgents->humans_no; i++) {
    humanRawPositions[i].positionAv  = FALSE;
    humanRawPositions[i].headAv      = FALSE;
    humanRawPositions[i].rightHandAv = FALSE;
    humanRawPositions[i].leftHandAv  = FALSE;
  }

#if defined(JIDO) || defined(HRP2)
  /* INPUT = MOTION CAPTURE */
  if(SDI_F->state.mocapPosterAv) {
    if (posterRead(gMocapPosterID, 0, &mocapObjects, sizeof(mocapObjects)) != sizeof(mocapObjects)) {
      *report = S_spark_POSTER_NOT_READ;
      return END;
    }
  
    mocap_translate_markers_pos(&mocapObjects, SPARK_MOCAP_FRAME_TRANSLATION_X, SPARK_MOCAP_FRAME_TRANSLATION_Y);  
    
    for(i=0; i<mocapObjects.nbodies; i++) {  // FOR EACH TRACKED THING IN MOCAP
      // For the moment 0 is the helmet, 1 is the right hand of this peson,
      // 2 is the glasses, 3 is the right hand of this person
      // TODO: Choose processing method according to the names in mocap poster
      
      if(i==0) { // HELMET -> HEAD 1st HUMAN
	if(mocapObjects.objects[i].status == 0) { // Status 0 means everything is ok
	  	  
	  // Human 0 is the one with the helmet 
	  if(mocap_compute_neck_position_from_helmet(&mocapObjects.objects[i], 
						     &humanRawPositions[0].head.x,
						     &humanRawPositions[0].head.y,
						     &humanRawPositions[0].head.z,
						     &humanRawPositions[0].head.yaw,
						     &humanRawPositions[0].head.pitch)) {
	    headOrientationAvailable = TRUE;
	    humanRawPositions[0].positionAv = TRUE;
	    humanRawPositions[0].headAv     = TRUE;
	  }
	}
      }
      if(i==1) { // GLOVES -> Right hand 1st HUMAN
	if(mocapObjects.objects[i].status == 0) { // Status 0 means everything is ok
	  
	  if(mocap_compute_palm_position_from_glove(&mocapObjects.objects[i],
						    &humanRawPositions[0].rightHand.x,
						    &humanRawPositions[0].rightHand.y,
						    &humanRawPositions[0].rightHand.z)) {
	    humanRawPositions[0].positionAv  = TRUE;
	    humanRawPositions[0].rightHandAv = TRUE;
	  }
	}
      }
      if(i==2) { // GLASSES -> HEAD 2nd HUMAN
	if(mocapObjects.objects[i].status == 0) { // Status 0 means everything is ok
	  
	  // Human 2 is the one wearing glasses
	  if(mocap_compute_neck_position_from_glasses(&mocapObjects.objects[i], 
						      &humanRawPositions[1].head.x,
						      &humanRawPositions[1].head.y,
						      &humanRawPositions[1].head.z,
						      &humanRawPositions[1].head.yaw,
						      &humanRawPositions[1].head.pitch)) {	    
	    headOrientationAvailable = TRUE;
	    humanRawPositions[1].positionAv = TRUE;
	    humanRawPositions[1].headAv     = TRUE;
	  }
	}
      }
      
      if(i==3) { // GLOVES -> Left Hand 2nd HUMAN
	if(mocapObjects.objects[i].status == 0) { // Status 0 means everything is ok
	  
	  if(mocap_compute_palm_position_from_glove(&mocapObjects.objects[i],
						    &humanRawPositions[1].leftHand.x,
						    &humanRawPositions[1].leftHand.y,
						    &humanRawPositions[1].leftHand.z)) {
	    humanRawPositions[1].positionAv  = TRUE;
	    humanRawPositions[1].leftHandAv  = TRUE;
	  }	  
	}
      }
    }  
  }
  
#ifdef GEST_HUMAN_TRACK
  /* INPUT = GEST */
  if(SDI_F->state.handPosterAv) {
    if (posterRead(gGestPosterID, 0, &gestPoster, sizeof(gestPoster)) != sizeof(gestPoster)) {
      *report = S_spark_POSTER_NOT_READ;
      return END;
    }
    if(gestPoster.pos.ETAT == TOUT_GENOM ||
       gestPoster.pos.ETAT == PASmAINdROITE_GENOM ||
       gestPoster.pos.ETAT == PAStETE_GENOM ||
       gestPoster.pos.ETAT == PAStETEnImAINdROITE_GENOM) {
      humanRawPositions[0].leftHandAv  = TRUE;
      humanRawPositions[0].positionAv  = TRUE;
    }
    if(gestPoster.pos.ETAT == TOUT_GENOM ||
       gestPoster.pos.ETAT == PASmAINgAUCHE_GENOM ||
       gestPoster.pos.ETAT == PAStETE_GENOM ||
       gestPoster.pos.ETAT == PAStETEnImAINgAUCHE_GENOM) {
      humanRawPositions[0].rightHandAv = TRUE;
      humanRawPositions[0].positionAv  = TRUE;
    }
    if(gestPoster.pos.ETAT == TOUT_GENOM ||
       gestPoster.pos.ETAT == PASmAINgAUCHE_GENOM ||
       gestPoster.pos.ETAT == PASmAINdROITE_GENOM ||
       gestPoster.pos.ETAT == AUCUNEmAIN_GENOM) {
      humanRawPositions[0].headAv     = TRUE;
      humanRawPositions[0].positionAv  = TRUE;
    } 
    if(gestPoster.HeadDetected)
      lookingAtRobot = TRUE;
    
    if(gestPoster.pos.ETAT == RIEN_GENOM) {
      /* gest doesn't find anything */
      return EXEC;
    }
    

      pomEulerVCompIn(&(cameraToOrigin), 
		    &(gestPoster.config.sensorToMain),
		    &(gestPoster.config.mainToOrigin));
    
      pomEulerVToT3D(&(cameraToOrigin), &gestSensorMatrix);
      t3dConvertTo(T3D_MATRIX, &gestSensorMatrix);

    
    if(humanRawPositions[0].rightHandAv) {
      humanRawPositions[0].rightHand.x = gestSensorMatrix.matrix.matrix[0] * gestPoster.pos.MDx/1000 +
                                         gestSensorMatrix.matrix.matrix[1] * gestPoster.pos.MDy/1000 +
                                         gestSensorMatrix.matrix.matrix[2] * gestPoster.pos.MDz/1000 +
                                         gestSensorMatrix.matrix.matrix[3];
      humanRawPositions[0].rightHand.y = gestSensorMatrix.matrix.matrix[4] * gestPoster.pos.MDx/1000 +
                                         gestSensorMatrix.matrix.matrix[5] * gestPoster.pos.MDy/1000 +
                                         gestSensorMatrix.matrix.matrix[6] * gestPoster.pos.MDz/1000 +
                                         gestSensorMatrix.matrix.matrix[7];
      humanRawPositions[0].rightHand.z = gestSensorMatrix.matrix.matrix[8] * gestPoster.pos.MDx/1000 +
                                         gestSensorMatrix.matrix.matrix[9] * gestPoster.pos.MDy/1000 +
                                         gestSensorMatrix.matrix.matrix[10] * gestPoster.pos.MDz/1000 +
                                      	 gestSensorMatrix.matrix.matrix[11];
    }
    
    /* Left hand */
    if(humanRawPositions[0].leftHandAv) {
      humanRawPositions[0].leftHand.x = gestSensorMatrix.matrix.matrix[0] * gestPoster.pos.MGx/1000 +
	                                gestSensorMatrix.matrix.matrix[1] * gestPoster.pos.MGy/1000 +
                                       	gestSensorMatrix.matrix.matrix[2] * gestPoster.pos.MGz/1000 +
                                       	gestSensorMatrix.matrix.matrix[3];
      humanRawPositions[0].leftHand.y = gestSensorMatrix.matrix.matrix[4] * gestPoster.pos.MGx/1000 +
                                      	gestSensorMatrix.matrix.matrix[5] * gestPoster.pos.MGy/1000 +
                                      	gestSensorMatrix.matrix.matrix[6] * gestPoster.pos.MGz/1000 +
                                    	gestSensorMatrix.matrix.matrix[7];
      humanRawPositions[0].leftHand.z = gestSensorMatrix.matrix.matrix[8] * gestPoster.pos.MGx/1000 +
                                       	gestSensorMatrix.matrix.matrix[9] * gestPoster.pos.MGy/1000 +
                                    	gestSensorMatrix.matrix.matrix[10] * gestPoster.pos.MGz/1000 +
                                   	gestSensorMatrix.matrix.matrix[11];
    }
    
    /* Head */
    if(humanRawPositions[0].headAv) {
      humanRawPositions[0].head.x = gestSensorMatrix.matrix.matrix[0] * gestPoster.pos.Tx/1000 +
	                            gestSensorMatrix.matrix.matrix[1] * gestPoster.pos.Ty/1000 +
                                    gestSensorMatrix.matrix.matrix[2] * gestPoster.pos.Tz/1000 +
                                    gestSensorMatrix.matrix.matrix[3];
      humanRawPositions[0].head.y = gestSensorMatrix.matrix.matrix[4] * gestPoster.pos.Tx/1000 +
                                    gestSensorMatrix.matrix.matrix[5] * gestPoster.pos.Ty/1000 +
                                    gestSensorMatrix.matrix.matrix[6] * gestPoster.pos.Tz/1000 +
                                    gestSensorMatrix.matrix.matrix[7];
      humanRawPositions[0].head.z = gestSensorMatrix.matrix.matrix[8] * gestPoster.pos.Tx/1000 +
                                    gestSensorMatrix.matrix.matrix[9] * gestPoster.pos.Ty/1000 +
                                    gestSensorMatrix.matrix.matrix[10] * gestPoster.pos.Tz/1000 +
                                    gestSensorMatrix.matrix.matrix[11];
    }
  }
#endif
#endif

#if defined(JIDO)
  /* INPUT = NIUT */
  if(SDI_F->state.niutPosterAv) {
    if (posterRead(gNiutPosterID, 0, &niutPoster, sizeof(niutPoster)) != sizeof(niutPoster)) {
      *report = S_spark_POSTER_NOT_READ;
      return END;
    }	

    p3d_matrix4 mKinectToOrigin;

    if(!isKinectOnTable)
      {
	if(SDI_F->state.kinectPosterAv) {
	  if (posterRead(gKinectPosterID, 0, &kinectCfg, sizeof(kinectCfg)) != sizeof(kinectCfg)) {
	    *report =  S_spark_POSTER_NOT_READ;
	    return END;
	  }
	}
  
	pomEulerVToT3D(&(kinectCfg.sensorToMain), &kinectToRobot);
	pomEulerVToT3D(&(kinectCfg.mainToOrigin), &robotToOrigin);

	kinectToRobot.euler.flags |= T3D_ALLOW_CONVERSION;
	robotToOrigin.euler.flags |= T3D_ALLOW_CONVERSION;

	t3dConvertTo(T3D_EULER, &kinectToRobot);
	t3dConvertTo(T3D_EULER, &robotToOrigin);

	//  t3dFPrint(stdout, &kinectToRobot );
	//  t3dFPrint(stdout, &robotToOrigin );

	T3D* kinectToOrigin = t3dComp( &kinectToRobot, &robotToOrigin);

	t3dConvertTo(T3D_MATRIX, kinectToOrigin );

	for(unsigned int i=0; i<4; i++) {
	  for(unsigned int j=0; j<4; j++) {
	    mKinectToOrigin[i][j] = kinectToOrigin->matrix[i*4+j];
	  }
	}
      }
    else
      {

	double baseX = 	gRobot->robotPt->joints[1]->abs_pos[0][3];
	double baseY = 	gRobot->robotPt->joints[1]->abs_pos[1][3];
	double baseZ = 1.8;

	mKinectToOrigin[0][0] = 0.0;  mKinectToOrigin[0][1] = 0.0;  mKinectToOrigin[0][2] = 1.0;  mKinectToOrigin[0][3] = baseX;  
	mKinectToOrigin[1][0] = 1.0;  mKinectToOrigin[1][1] = 0.0;  mKinectToOrigin[1][2] = 0.0;  mKinectToOrigin[1][3] = baseY;  
	mKinectToOrigin[2][0] = 0.0;  mKinectToOrigin[2][1] = 1.0;  mKinectToOrigin[2][2] = 0.0;  mKinectToOrigin[2][3] = baseZ;  
	mKinectToOrigin[3][0] = 0.0;  mKinectToOrigin[3][1] = 0.0;  mKinectToOrigin[3][3] = 0.0;  mKinectToOrigin[3][3] = 1.0;  
      }

  //  printf("x = %f, y = %f, z = %f " ,  mKinectToOrigin[0][3], mKinectToOrigin[1][3], mKinectToOrigin[2][3] );

  // copy head platine position to config
  //  Pt2ConfigPt(platineFrame.euler.euler[0], platineFrame.euler.euler[1], gRRRobotQ);

    unsigned int HumanId = 16;

    for(unsigned int i=0; i<16;i++)
    {
      if( niutPoster.users[i].state == NIUT_TRACKING )
	{
	  HumanId = i;
	  break;
	}
    }

    if( HumanId != 16 )
    {
      p3d_rob* robot = p3d_get_robot_by_name_containing("HUMAN");
      
      kinectData data;

      niutGenPosToVect(mKinectToOrigin , data.TORSO,         niutPoster.users[HumanId].skeleton.joint[NIUT_TORSO].position);
      niutGenPosToVect(mKinectToOrigin , data.NECK,          niutPoster.users[HumanId].skeleton.joint[NIUT_NECK].position);
      niutGenPosToVect(mKinectToOrigin , data.HEAD,          niutPoster.users[HumanId].skeleton.joint[NIUT_HEAD].position);

      niutGenPosToVect(mKinectToOrigin , data.SHOULDER_RIGHT, niutPoster.users[HumanId].skeleton.joint[NIUT_RIGHT_SHOULDER].position);
      niutGenPosToVect(mKinectToOrigin , data.ELBOW_RIGHT,    niutPoster.users[HumanId].skeleton.joint[NIUT_RIGHT_ELBOW].position);
      niutGenPosToVect(mKinectToOrigin , data.HAND_RIGHT,     niutPoster.users[HumanId].skeleton.joint[NIUT_RIGHT_HAND].position);

      niutGenPosToVect(mKinectToOrigin , data.SHOULDER_LEFT,  niutPoster.users[HumanId].skeleton.joint[NIUT_LEFT_SHOULDER].position);
      niutGenPosToVect(mKinectToOrigin , data.ELBOW_LEFT,     niutPoster.users[HumanId].skeleton.joint[NIUT_LEFT_ELBOW].position);
      niutGenPosToVect(mKinectToOrigin , data.HAND_LEFT,      niutPoster.users[HumanId].skeleton.joint[NIUT_LEFT_HAND].position);
     
      niutGenPosToVect(mKinectToOrigin , data.HIP_RIGHT,      niutPoster.users[HumanId].skeleton.joint[NIUT_RIGHT_HIP].position);
      niutGenPosToVect(mKinectToOrigin , data.KNEE_RIGHT,     niutPoster.users[HumanId].skeleton.joint[NIUT_RIGHT_KNEE].position);
      niutGenPosToVect(mKinectToOrigin , data.FOOT_RIGHT,     niutPoster.users[HumanId].skeleton.joint[NIUT_RIGHT_FOOT].position);

      niutGenPosToVect(mKinectToOrigin, data.HIP_LEFT,        niutPoster.users[HumanId].skeleton.joint[NIUT_LEFT_HIP].position);
      niutGenPosToVect(mKinectToOrigin, data.KNEE_LEFT,       niutPoster.users[HumanId].skeleton.joint[NIUT_LEFT_KNEE].position);
      niutGenPosToVect(mKinectToOrigin, data.FOOT_LEFT,       niutPoster.users[HumanId].skeleton.joint[NIUT_LEFT_FOOT].position);

      hri_store_kinect_model( data );
      
      configPt q = hri_get_configuration_from_kinect_data( robot, data );
      p3d_set_and_update_this_robot_conf( robot, q );
      p3d_destroy_config( robot, q );
    }
  }
#endif

  /* INPUT = A FULL BODY CONFIGURATION */
  if(SDI_F->state.humanConfPosterAv) {
    // ONLY FOR 1 HUMAN
    // We take the whole configuration and load directly to move3d
    if (posterRead(gHumanConfPosterID, 0, &humanConf, sizeof(humanConf)) != sizeof(humanConf)) {
	  printf("Unable to read the poster of human configuration :-/");
      *report = S_spark_POSTER_NOT_READ;
      return END;
    }
    if(humanConf.changed) {

      for(i=0;i<40;i++) {
			gRHHumanQ[i] = humanConf.dof[i];
      }
      for(; i<gAgents->humans[0]->robotPt->nb_dof; i++) {
	gRHHumanQ[i] = 0;
      }
      
      p3d_set_and_update_this_robot_conf(gAgents->humans[0]->robotPt, gRHHumanQ);
      SDI_F->updateObjectsStates = TRUE;
    }
    
    return EXEC; 
  }
  /* INPUT = BODY PARTS */
  if(SDI_F->state.bodyPartsPosterAv) {
    if (posterRead(gBodyPartsPosterID, 0, &humanBodyParts, sizeof(humanBodyParts)) != sizeof(humanBodyParts)) {
      *report = S_spark_POSTER_NOT_READ;
      return END;
    }
    for(i=0; i<humanBodyParts.humanNb; i++) {
      for(hIdx=0; hIdx<gAgents->humans_no; hIdx++) {
	if(!strcmp(humanBodyParts.human[i].name.name, gAgents->humans[hIdx]->robotPt->name)) {
	  if(humanBodyParts.human[i].headAv) {
	    panDofIdx  = gAgents->humans[hIdx]->robotPt->joints[gAgents->humans[hIdx]->perspective->pan_jnt_idx]->index_dof;
	    tiltDofIdx = gAgents->humans[hIdx]->robotPt->joints[gAgents->humans[hIdx]->perspective->tilt_jnt_idx]->index_dof;
	    
	    p3d_get_robot_config_into(gAgents->humans[hIdx]->robotPt, &gRHHumanQ);
	    
	    gRHHumanQ[6] = humanBodyParts.human[i].head.x;
	    gRHHumanQ[7] = humanBodyParts.human[i].head.y;

	    gRHHumanQ[panDofIdx] = humanBodyParts.human[i].head.roll;

	    hri_agent_compute_posture(gAgents->humans[hIdx],  humanBodyParts.human[i].head.z-0.15, STANDING , gRHHumanQ);
	    
	    p3d_set_and_update_this_robot_conf(gAgents->humans[0]->robotPt, gRHHumanQ);
	   
	    if(humanBodyParts.human[i].rightHandAv) {
	      hri_agent_load_default_arm_posture(gAgents->humans[hIdx], gRHHumanQ);
	      hri_agent_single_task_manip_move(gAgents->humans[hIdx], GIK_RAREACH, &tCoord, 0.01, &gRHHumanQ);
	      p3d_set_and_update_this_robot_conf(gAgents->humans[0]->robotPt, gRHHumanQ);
	    }
	    if(humanBodyParts.human[i].leftHandAv) {
	      hri_agent_load_default_arm_posture(gAgents->humans[hIdx], gRHHumanQ);
	      hri_agent_single_task_manip_move(gAgents->humans[hIdx], GIK_LAREACH, &tCoord, 0.01, &gRHHumanQ);
	      p3d_set_and_update_this_robot_conf(gAgents->humans[0]->robotPt, gRHHumanQ);
	    }

	    SDI_F->updateObjectsStates = TRUE;
	  }
	}
      }
    }
    return EXEC; 
  }

  /*------ END INPUTS --------*/
  
  /* Updating HUMAN */
  
  for(hIdx=0; hIdx<gAgents->humans_no; hIdx++) {
    
    if(humanRawPositions[hIdx].positionAv) {
      
      gAgents->humans[hIdx]->is_present = TRUE;

      panDofIdx  = gAgents->humans[hIdx]->robotPt->joints[gAgents->humans[hIdx]->perspective->pan_jnt_idx]->index_dof;
      tiltDofIdx = gAgents->humans[hIdx]->robotPt->joints[gAgents->humans[hIdx]->perspective->tilt_jnt_idx]->index_dof;
      
      p3d_get_robot_config_into(gAgents->humans[hIdx]->robotPt, &gRHHumanQ);
  
      GIK_VIS = 500; // To disable the interface refresh when computing reach posture
      
      // First we place the head of the human. Then we compute his body posture.
      // Then we make reaching motions with arms
      // For the moment reaching motion should NOT involve torso since it will change the body posture
      
      if(humanRawPositions[hIdx].headAv) {

	gAgents->humans[hIdx]->head[0]->is_present = TRUE; // A human has only one head
	
	updateDistDiff = D3D(gHumanPrevPositions[hIdx].history[0].head.x,
			     gHumanPrevPositions[hIdx].history[0].head.y,
			     gHumanPrevPositions[hIdx].history[0].head.z,
			     humanRawPositions[hIdx].head.x,
			     humanRawPositions[hIdx].head.y,
			     humanRawPositions[hIdx].head.z );

	updateAngleDiff = D2D(gHumanPrevPositions[hIdx].history[0].head.yaw,
			      gHumanPrevPositions[hIdx].history[0].head.pitch,
			      humanRawPositions[hIdx].head.yaw,
			      humanRawPositions[hIdx].head.pitch);
	
	if(updateDistDiff > 0.4 ) {
	  //There is a big jump, might come from a sensor error. Need to perceive it several times before accepting it
	  gHumanHeadJumpCounter[hIdx]++;
	} else {
	  gHumanHeadJumpCounter[hIdx] = 0;
	}
	if (gHumanHeadJumpCounter[hIdx] > 5) {
	  gHumanHeadJumpCounter[hIdx] = 0;
	}
	// update sensor readings unless there is a jump
	if(gHumanHeadJumpCounter[hIdx] == 0) {

	  // update posture if change is significant or state changed (stopped moving)
	  if(updateDistDiff>SPARK_HUMAN_DEFAULT_UPDATE_DIST  || updateAngleDiff>SPARK_HUMAN_DEFAULT_UPDATE_ANGLE) {	  
	    
	    spark_push_position_to_history(gHumanPrevPositions, hIdx, &humanRawPositions[hIdx].head, NULL, NULL);
	    
	    //state = spark_compute_human_state(gAgents->humans[hIdx], gHumanPrevPositions);
	    if(humanRawPositions[hIdx].head.z > SPARK_HUMAN_DEFAULT_SITTING_HEAD_HEIGHT)
	      state = STANDING;
	    else
	      state = SITTING;
	    
	    gAgents->robots[hIdx]->actual_state = state;

	    gRHHumanQ[6] = humanRawPositions[hIdx].head.x;
	    gRHHumanQ[7] = humanRawPositions[hIdx].head.y;
	    
	    if (lookingAtRobot) {
	      // if human is looking at the robot, we change his orientation towards the robot
	      gRHHumanQ[11] =  atan2(gAgents->robots[hIdx]->robotPt->joints[1]->dof_data[1].v-gRHHumanQ[7],
				     gAgents->robots[hIdx]->robotPt->joints[1]->dof_data[0].v-gRHHumanQ[6]);
	    }
	    if(headOrientationAvailable) {
	      if(ABS(gRHHumanQ[11]-humanRawPositions[hIdx].head.yaw) > M_PI_2*0.7) {
		gRHHumanQ[11] = humanRawPositions[hIdx].head.yaw; 
		gRHHumanQ[panDofIdx] = 0;
	      }
	      else {
		gRHHumanQ[panDofIdx] = humanRawPositions[hIdx].head.yaw - gRHHumanQ[11];
	      }
	      gRHHumanQ[tiltDofIdx] = humanRawPositions[hIdx].head.pitch;
	      
	    }
	    
	    // HEAD contains the position of the center of the head. Compute_posture position requires the position of the neck
	    // Thus for the moment we decrease the height by a constant (0.15).
	    // TODO: compute this value automatically from the model
	    hri_agent_compute_posture(gAgents->humans[hIdx], humanRawPositions[hIdx].head.z-0.15, state, gRHHumanQ);
	    
	    // write new human position to poster
	    //SDI_F->HumanPositions.human[hIdx].pos.x  = gRHHumanQ[6];
	    //SDI_F->HumanPositions.human[hIdx].pos.y  = gRHHumanQ[7];
	    //SDI_F->HumanPositions.human[hIdx].pos.th = gRHHumanQ[11];
	    //SDI_F->HumanPositions.human[hIdx].state  = gAgents->humans[hIdx]->actual_state;
	    
	    headIsUpdated = TRUE;
	    updateHumanPosture = TRUE;
	    
	    //printf("UPDATE BY HEAD\n");
	  }
	}
      }
      if(humanRawPositions[hIdx].leftHandAv || headIsUpdated) { // if head is updated, we have to recompute hand positions in anyways
	if(headIsUpdated) {
	  if(humanRawPositions[hIdx].leftHandAv) {
	    
	    gAgents->humans[hIdx]->hand[0]->is_present = TRUE; // 0 is left hand...

	    spark_push_position_to_history(gHumanPrevPositions, hIdx, NULL, &humanRawPositions[hIdx].leftHand, NULL);
	    
	    tCoord[0] = humanRawPositions[hIdx].leftHand.x;      
	    tCoord[1] = humanRawPositions[hIdx].leftHand.y;     
	    tCoord[2] = humanRawPositions[hIdx].leftHand.z;
	  }
	  else {
	    tCoord[0] = gHumanPrevPositions[hIdx].history[0].lhand.x;      
	    tCoord[1] = gHumanPrevPositions[hIdx].history[0].lhand.y;     
	    tCoord[2] = gHumanPrevPositions[hIdx].history[0].lhand.z;
	  }
	  
	  if( gHumanLHandReachFailCounter[hIdx] > 3 ) { // hand update failed 5 times in a row. We put the human into a default posture
	    hri_agent_load_default_arm_posture(gAgents->humans[hIdx], gRHHumanQ);
	    gHumanLHandReachFailCounter[hIdx] = 0;
	  }
	  p3d_set_and_update_this_robot_conf(gAgents->humans[hIdx]->robotPt, gRHHumanQ); 
	  // Next funtion takes current human pos by default. That's why we have to update human config
	  
	  if(!hri_agent_single_task_manip_move(gAgents->humans[hIdx], GIK_LAWREACH, &tCoord, 0.01, &gRHHumanQ))	    
	    gHumanLHandReachFailCounter[hIdx]++;
	  else
	    gHumanLHandReachFailCounter[hIdx] = 0;
	  
	  updateHumanPosture = TRUE;
	}
	else {
	  gAgents->humans[hIdx]->hand[0]->is_present = TRUE; // 0 is left hand...
	  
	  updateDistDiff = D3D(humanRawPositions[hIdx].leftHand.x,
			       humanRawPositions[hIdx].leftHand.y,
			       humanRawPositions[hIdx].leftHand.z,
			       gHumanPrevPositions[hIdx].history[0].lhand.x,
			       gHumanPrevPositions[hIdx].history[0].lhand.y,
			       gHumanPrevPositions[hIdx].history[0].lhand.z);
	  
	  if(updateDistDiff > 0.3 ) {
	    //There is a big jump, might come from a sensor error. Need to perceive it several times before accepting it
	    //printf("BIG JUMP Lhand no: %d\n", gHumanLHandJumpCounter[hIdx]);
	    gHumanLHandJumpCounter[hIdx]++;
	  } else {
	    gHumanLHandJumpCounter[hIdx] = 0;
	  }
	  if (gHumanLHandJumpCounter[hIdx] > 5) {
	    gHumanLHandJumpCounter[hIdx] = 0;
	  }
	  // update sensor readings unless there is a jump
	  if(gHumanLHandJumpCounter[hIdx] == 0) {
	    
	    if(( updateDistDiff > SPARK_HUMAN_DEFAULT_HAND_UPDATE_DIST) && (updateDistDiff < 0.4)) {
	      
	      spark_push_position_to_history(gHumanPrevPositions, hIdx, NULL, &humanRawPositions[hIdx].leftHand, NULL);
	      
	      tCoord[0] = humanRawPositions[hIdx].leftHand.x;      
	      tCoord[1] = humanRawPositions[hIdx].leftHand.y;    
	      tCoord[2] = humanRawPositions[hIdx].leftHand.z;
	      
	      p3d_set_and_update_this_robot_conf(gAgents->humans[hIdx]->robotPt, gRHHumanQ); // Next funtion takes current human pos by default. That's why we have to update human config
	      hri_agent_load_default_arm_posture(gAgents->humans[hIdx], gRHHumanQ);
	      hri_agent_single_task_manip_move(gAgents->humans[hIdx], GIK_LAWREACH, &tCoord, 0.01, &gRHHumanQ);
	      
	      updateHumanPosture = TRUE;
	    }
	  }
	}
      }
      if(humanRawPositions[hIdx].rightHandAv || headIsUpdated) { // if head is updated, we have to recompute hand positions in anyways
	if(headIsUpdated) {
	  if(humanRawPositions[hIdx].rightHandAv) {
	    gAgents->humans[hIdx]->hand[1]->is_present = TRUE; // 1 is right hand...

	    spark_push_position_to_history(gHumanPrevPositions, hIdx, NULL, &humanRawPositions[hIdx].rightHand, NULL);
	    
	    tCoord[0] = humanRawPositions[hIdx].rightHand.x;      
	    tCoord[1] = humanRawPositions[hIdx].rightHand.y;   
	    tCoord[2] = humanRawPositions[hIdx].rightHand.z;
	  }
	  else {
	    tCoord[0] = gHumanPrevPositions[hIdx].history[0].rhand.x;      
	    tCoord[1] = gHumanPrevPositions[hIdx].history[0].rhand.y;     
	    tCoord[2] = gHumanPrevPositions[hIdx].history[0].rhand.z;
	  }
	  
	  if( gHumanRHandReachFailCounter[hIdx] > 3 ) { // hand update failed 5 times in a row. We put the human into a default posture
	    hri_agent_load_default_arm_posture(gAgents->humans[hIdx], gRHHumanQ);
	    gHumanRHandReachFailCounter[hIdx] = 0;
	  }
	  p3d_set_and_update_this_robot_conf(gAgents->humans[hIdx]->robotPt, gRHHumanQ); 
	  // Next funtion takes current human pos by default. That's why we have to update human config
	  
	  if(!hri_agent_single_task_manip_move(gAgents->humans[hIdx], GIK_RAWREACH, &tCoord, 0.01, &gRHHumanQ))	    
	    gHumanRHandReachFailCounter[hIdx]++;
	  else
	    gHumanRHandReachFailCounter[hIdx] = 0;
	  
	  updateHumanPosture = TRUE;
	}
	else {
	  gAgents->humans[hIdx]->hand[1]->is_present = TRUE; // 1 is right hand...
	  
	  updateDistDiff = D3D(humanRawPositions[hIdx].rightHand.x,
			       humanRawPositions[hIdx].rightHand.y,
			       humanRawPositions[hIdx].rightHand.z,
			       gHumanPrevPositions[hIdx].history[0].rhand.x,
			       gHumanPrevPositions[hIdx].history[0].rhand.y,
			       gHumanPrevPositions[hIdx].history[0].rhand.z);
	  
	  if(updateDistDiff > 0.3 ) {
	    //There is a big jump, might come from a sensor error. Need to perceive it several times before accepting it
	    //printf("BIG JUMP Lhand no: %d\n", gHumanRHandJumpCounter[hIdx]);
	    gHumanRHandJumpCounter[hIdx]++;
	  } else {
	    gHumanRHandJumpCounter[hIdx] = 0;
	  }
	  if (gHumanRHandJumpCounter[hIdx] > 5) {
	    gHumanRHandJumpCounter[hIdx] = 0;
	  }
	  
	  if( (updateDistDiff > SPARK_HUMAN_DEFAULT_HAND_UPDATE_DIST) && (updateDistDiff < 0.4) ) {
	    
	    spark_push_position_to_history(gHumanPrevPositions, hIdx, NULL, NULL, &humanRawPositions[hIdx].rightHand);
	    
	    tCoord[0] = humanRawPositions[hIdx].rightHand.x;      
	    tCoord[1] = humanRawPositions[hIdx].rightHand.y;    
	    tCoord[2] = humanRawPositions[hIdx].rightHand.z;
	    
	    p3d_set_and_update_this_robot_conf(gAgents->humans[hIdx]->robotPt, gRHHumanQ); // Next funtion takes current human pos by default. 
	                                                                                   // That's why we have to update human config
	    hri_agent_load_default_arm_posture(gAgents->humans[hIdx], gRHHumanQ);
	    hri_agent_single_task_manip_move(gAgents->humans[hIdx], GIK_RAWREACH, &tCoord, 0.01, &gRHHumanQ);
	    
	    updateHumanPosture = TRUE;
	  }
	}
      }
      
      if(updateHumanPosture) {
	//printf("Updating human to %f %f %f\n",gRHHumanQ[6],gRHHumanQ[7],gRHHumanQ[8]);
	p3d_set_and_update_this_robot_conf(gAgents->humans[hIdx]->robotPt, gRHHumanQ);
	place_robot_conf_to_environment_poster(gAgents->humans[hIdx]->robotPt, gRHHumanQ);
	SDI_F->updateObjectsStates = TRUE;
	printf("human moved MOVED\n");
	
        // Computes and exposes the minimal distance and cost
        // between the robot and all the humans in the scene
        SDI_F->distToHuman.costDistRobToHum = hri_distance_cost( gAgents, SDI_F->distToHuman.minDistRobToHum );
	sparkDistanceToHumandistToHumanPosterWrite(SPARK_DISTANCETOHUMAN_POSTER_ID, &SDI_F->distToHuman);
  
	if(gAgents->humans[hIdx]->perspective->enable_visible_objects_draw) {
	  hri_compute_agent_sees(gAgents->humans[0], 50, FALSE, FALSE);
	}
      }
    }
  }
  
  publish_environment_poster();
  //if(SDI_F->state.mightabilitiesEnabled)
  //  execute_Mightability_Map_functions();
  
  return EXEC; 
}


/* sparkReadHumansEnd  -  codel END of ReadHumans
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkReadHumansEnd(SPARK_INPUT_POSTER *humanPoster, int *report)
{  
  MY_FREE(gRHHumanQ, double, gAgents->humans[0]->robotPt->nb_dof); /* FREE */
  MY_FREE(gHumanPrevPositions, SPARK_AGENT_HISTORY, 2); /* FREE */
  printf("Human Acquisition stopped.\n");
  return ETHER;
}


/* sparkReadHumansInter  -  codel INTER of ReadHumans
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkReadHumansInter(SPARK_INPUT_POSTER *humanPoster, int *report)
{ 
  MY_FREE(gRHHumanQ, double, gAgents->humans[0]->robotPt->nb_dof); /* FREE */
  MY_FREE(gHumanPrevPositions, SPARK_AGENT_HISTORY, 2); /* FREE */
  printf("Human Acquisition interrupted.\n");
  return ETHER;
}

/* ------------------------------------------------------------- */
/* ----------------------- GET JOINT POS ----------------------- */
/* ------------------------------------------------------------- */

/*----------------------------------------------------------------
 * GetJointAbsPose
 *
 * Description: 
 *
 * Reports:      OK
 *              S_spark_ROBOT_NOT_FOUND
 *              S_spark_JOINT_NOT_FOUND
 *              S_spark_NOT_INITIALIZED
 */

/* sparkGetJointAbsPoseMain  -  codel EXEC of GetJointAbsPose
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkGetJointAbsPoseMain(SPARK_JOINT_ABS_POS_PARAM *jointAbsPoseParam, POM_EULER *jointAbsPose, int *report)
{

  if(gEnv == NULL) {
    printf("SPARK is not properly initialized.\n");
    *report = S_spark_NOT_INITIALIZED;
    return ETHER;
  }

  p3d_rob* rob = p3d_get_robot_by_name(jointAbsPoseParam->robotName.name);

  if( rob == NULL ) {
    *report = S_spark_ROBOT_NOT_FOUND;
    return ETHER;
  }
  
  p3d_jnt* jnt = p3d_get_robot_jnt_by_name( rob, jointAbsPoseParam->jointName.name);

  if( jnt == NULL ) {
    *report = S_spark_JOINT_NOT_FOUND;
    return ETHER;
  }

  // Convert Move3D abs_pos to Euler (Yaw,Pitch,Roll)
  T3D* jointToOrigin = t3dNew( T3D_MATRIX, T3D_ALLOW_CONVERSION );

  for(unsigned int i=0; i<4; i++) {
    for(unsigned int j=0; j<4; j++) {
      jointToOrigin->matrix[4*i+j] = jnt->abs_pos[i][j];
      }
   }

  t3dConvertTo( T3D_BRYAN, jointToOrigin );

  jointAbsPose->x      = jointToOrigin->euler.x;
  jointAbsPose->y      = jointToOrigin->euler.y;
  jointAbsPose->z      = jointToOrigin->euler.z;
  jointAbsPose->yaw    = jointToOrigin->euler.yaw;
  jointAbsPose->pitch  = jointToOrigin->euler.pitch;
  jointAbsPose->roll   = jointToOrigin->euler.roll;


  delete jointToOrigin;

  //  p3d_mat4ExtractPosReverseOrder(M,
  //				 &(jointAbsPose->x), &jointAbsPose->y, jointAbsPose->z,
  //				 &(jointAbsPose->roll), &(jointAbsPos->pitch),&(

  return ETHER;
}

/* ------------------------------------------------------------- */
/* ----------------------- PLACE OBJECT ------------------------ */
/* ------------------------------------------------------------- */

/* sparkPlaceObjectMain  -  codel EXEC of PlaceObject
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkPlaceObjectMain(SPARK_M3D_OBJECT *addedObject, int *report)
{
  configPt robotQ;
  int i,e_j;
  p3d_rob * rob = NULL;

  
  for(i=0; i<gEnv->nr; i++) {
    if( strcasestr(gEnv->robot[i]->name, addedObject->name.name) ) {
      rob = gEnv->robot[i];
      break;
    }
  }
  
  // update entity is_present.
  for(e_j=0; e_j<gEntities->entities_nb; e_j++) {
    if (!strcmp(gEnv->robot[i]->name, gEntities->entities[e_j]->robotPt->name)){
      gEntities->entities[e_j]->is_present = TRUE;
      break;
    }
  }


  if( rob == NULL ) {
    *report = S_spark_OBJECT_NOT_FOUND;
    return ETHER;
  }
  
  robotQ = MY_ALLOC(double, rob->nb_dof); /* ALLOC */
  
  p3d_get_robot_config_into(rob, &robotQ);
  
  robotQ[6]  = addedObject->coords.x;
  robotQ[7]  = addedObject->coords.y;
  robotQ[8]  = addedObject->coords.z;
  robotQ[9]  = DTOR(addedObject->coords.yaw);
  robotQ[10] = DTOR(addedObject->coords.pitch);
  robotQ[11] = DTOR(addedObject->coords.roll);
  
  p3d_set_and_update_this_robot_conf(rob, robotQ);
  
  if(place_robot_conf_to_environment_poster(rob, robotQ))
    publish_environment_poster();
  
  MY_FREE(robotQ, double, rob->nb_dof); /* FREE */  
  
  SDI_F->updateObjectsStates = TRUE; 

  *report = OK;
  return ETHER;
}

/* ------------------------------------------------------------- */
/* ----------------------- PLACE OBJECT FOR SIMU --------------- */
/* ------------------------------------------------------------- */

/* sparkPlaceObjectMain  -  codel EXEC of PlaceObject
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkPlaceObjectForSimuMain(SPARK_M3D_OBJECT_SIMU_PARAMS *addedObjectSimuParams, int *report)
{
 
  SDI_F->addedObjectSimu.name = addedObjectSimuParams->name;
  SDI_F->addedObjectSimu.coords.x  = addedObjectSimuParams->coords.x;
  SDI_F->addedObjectSimu.coords.y  = addedObjectSimuParams->coords.y;
  SDI_F->addedObjectSimu.coords.z  = addedObjectSimuParams->coords.z;
  SDI_F->addedObjectSimu.coords.yaw = addedObjectSimuParams->coords.yaw;
  SDI_F->addedObjectSimu.coords.pitch = addedObjectSimuParams->coords.pitch;
  SDI_F->addedObjectSimu.coords.roll = addedObjectSimuParams->coords.roll;
  SDI_F->addedObjectSimu.hasDisappeared = addedObjectSimuParams->hasDisappeared;
  SDI_F->addedObjectSimu.isnew = 1;
  *report = OK;
  return ETHER;
}

/* ------------------------------------------------------------- */
/* ----------------------- Set Grasped Object ------------------ */
/* ------------------------------------------------------------- */

/* sparkPlaceObjectMain  -  codel EXEC of PlaceObject
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkSetGraspedObjectMain(SPARK_SET_GRASPED_OBJECT_PARAMS *graspedObjectParams, int *report)
{
 
  int e_j;
  HRI_AGENT * agent;
  
  agent = gAgents->all_agents[gAgents->source_agent_idx];

  hri_agent_is_grasping_obj(agent, graspedObjectParams->is_grasping_object ,graspedObjectParams->name.name ,graspedObjectParams->grasped_object_arm_id );
  agent->object_name = graspedObjectParams->name.name;
  agent->is_grasping_object = graspedObjectParams->is_grasping_object;
  agent->grasped_object_arm_id = graspedObjectParams->grasped_object_arm_id;

  ///  Put current entity robot_p3d in environement poster
  if(graspedObjectParams->is_grasping_object){
    for(e_j=0; e_j<gEntities->entities_nb; e_j++) {
      if (!strcmp(graspedObjectParams->name.name, gEntities->entities[e_j]->robotPt->name)){
	break;
      }
    }
    if(place_robot_p3d_rob_to_environment_poster(gEntities->entities[e_j]->robotPt))
      publish_environment_poster();
  }
  
  *report = OK;
  return ETHER;
}


/* ------------------------------------------------------------- */
/* ----------------------- PLACE AGENT ------------------------- */
/* ------------------------------------------------------------- */

/* sparkPlaceAgentMain  -  codel EXEC of PlaceAgent
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sparkPlaceAgentMain(SPARK_AGENT_POS *addedAgent, int *report)
{ 
  configPt agentQ;
  p3d_rob * rob;
  int i,e_j;
  int tiltDof, panDof;
  
  if(gAgents == NULL || gEnv == NULL) {
    *report = S_spark_NOT_INITIALIZED;
    return ETHER;
  }  
 
  for(i=0; i<gAgents->all_agents_no; i++) {
    if( strcasestr(gAgents->all_agents[i]->robotPt->name, addedAgent->name.name) ) {

      // update agent is present and entities of this agent is_present for situation assessment local testing.
      gAgents->all_agents[i]->is_present = TRUE;
      for(e_j=0; e_j<gEntities->entities_nb; e_j++) {	
	if( (gEntities->entities[e_j]->type == HRI_AGENT_PART) || (gEntities->entities[e_j]->type == HRI_ISAGENT) ) {
	  if( gAgents->all_agents[i] == gAgents->all_agents[gEntities->entities[e_j]->agent_idx] )
	    gEntities->entities[e_j]->is_present = TRUE;
	}
      }

      rob =  gAgents->all_agents[i]->robotPt;
      
      agentQ = MY_ALLOC(double,rob->nb_dof); /* ALLOC */
	
      p3d_get_robot_config_into(rob, &agentQ);
      
      agentQ[6]  = addedAgent->pos.x;       // x
      agentQ[7]  = addedAgent->pos.y;       // y
      agentQ[11] = DTOR(addedAgent->pos.z); // Rz
      
      panDof  = rob->joints[gAgents->all_agents[i]->perspective->pan_jnt_idx]->index_dof;
      tiltDof = rob->joints[gAgents->all_agents[i]->perspective->tilt_jnt_idx]->index_dof;
      
      agentQ[panDof]  = DTOR(addedAgent->pan);
      agentQ[tiltDof] = DTOR(addedAgent->tilt);
      
      hri_agent_load_default_arm_posture(gAgents->all_agents[i], agentQ);
      hri_agent_compute_state_posture(gAgents->all_agents[i], (int)addedAgent->state, agentQ);
      
      gAgents->all_agents[i]->actual_state = (int)addedAgent->state;

      p3d_set_and_update_this_robot_conf(rob, agentQ);
      
      // Computes and exposes the minimal distance and cost
      // between the robot and all the humans in the scene
      SDI_F->distToHuman.costDistRobToHum = hri_distance_cost( gAgents, SDI_F->distToHuman.minDistRobToHum );
      sparkDistanceToHumandistToHumanPosterWrite(SPARK_DISTANCETOHUMAN_POSTER_ID, &SDI_F->distToHuman);
       
      if(place_robot_conf_to_environment_poster(rob, agentQ))
	publish_environment_poster();

      MY_FREE(agentQ, double, rob->nb_dof); /* FREE */
	
      break;
    }
  }
  if(i == gAgents->all_agents_no) {
    *report = S_spark_AGENT_NOT_FOUND;
  }
  else {
    *report = OK;
    SDI_F->updateObjectsStates = TRUE; 
  }

  return ETHER;
}

